//p40fpga``+
#define P40_FMT "P40ECS:%s(): "
#define PCIE40_ECS_CLASS "pcie40_ecs"

#include "ecs.h"
#include "pcie40_ioctl.h"

//dg`wt.pcie40.fpga.ecs` The ECS submodule implements the system calls behind the BAR0 and BAR2 device files. The submodule is initialized and uninitialized at the same time as the main module, through the p40driver`pcie40_ecs_init` and p40driver`pcie40_ecs_exit` functions. Likewise the ECS-specific probing logic is encapsulated in p40fpga`pcie40_ecs_probe` and p40fpga`pcie40_ecs_remove` .

static void pcie40_ecs_set_drvdata(struct pci_dev *pdev, struct pcie40_ecs_state *state)
{
  struct pcie40_state *common = pci_get_drvdata(pdev);
  common->ecs_state = state;
  common->ecs_state->common = common;
}
static struct pcie40_ecs_state *pcie40_ecs_get_drvdata(struct pci_dev *pdev)
{
  struct pcie40_state *common = pci_get_drvdata(pdev);

  return common->ecs_state;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35))
static int ecs_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
#else
static long ecs_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
#endif //+`ecs_ioctl`
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 35))
  struct inode *inode = filp->f_inode;
#endif
  struct pcie40_ecs_state *state = filp->private_data;
  int bar ;
  int retval = 0 ;
  int i = 0 ;
  switch (iminor(inode))
  {
  case BAR0_CDEV_MINOR:
    bar = 0;
    break;
  case BAR2_CDEV_MINOR:
    bar = 2;
    break;
  default:
    return -EINVAL;
  }

  switch (cmd)
  {
    case P40_ECS_GET_BAR_SIZE: 
      pr_devel(P40_FMT "ECS BAR size is %lu\n", P40_PARM, state->common->bar_size[bar]);
      retval = state->common->bar_size[bar];
      break ;
    case P40_ECS_SLC_LOCK:
      retval = mutex_lock_interruptible( &(pcie40_slc_mutex[arg]) ) ;
      if ( 0 != retval ) printk(P40_WARN "problem with slow control lock\n", P40_PARM);
      break ;
    case P40_ECS_SLC_UNLOCK:
      mutex_unlock( &(pcie40_slc_mutex[arg]) ) ;
      retval = 0 ;
      break ;
    case P40_ECS_SLC_CHECKLOCK:
      while ( mutex_is_locked( &(pcie40_slc_mutex[arg]) ) ) {
        ++i ; 
        if ( i > 10000 ) { 
          printk(P40_WARN "force slow control lock release\n", P40_PARM);
          return -EBUSY ;
        }
      }
      retval = 0 ;
      break;
    default:
      printk(P40_WARN "invalid ioctl command\n", P40_PARM);
      return -EINVAL;
  }
  
  return retval ; 
}

//+`ecs_mmap`
static int ecs_mmap(struct file *filp, struct vm_area_struct *vma) //;?>
{
  int rc = 0;
  struct pcie40_ecs_state *state = filp->private_data;

  int bar;
  switch (iminor(filp->f_path.dentry->d_inode))
  {
  case BAR0_CDEV_MINOR:
    bar = 0;
    break;
  case BAR2_CDEV_MINOR:
    bar = 2;
    break;
  default:
    return -EINVAL;
  }
  vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

  pr_devel(P40_FMT "BAR %d VIRT=%lx PHYS=%lx SIZE=%lu\n", P40_PARM, bar,
           vma->vm_start, state->common->bar_start[bar], vma->vm_end - vma->vm_start);

  rc = io_remap_pfn_range(vma, vma->vm_start,
                          state->common->bar_start[bar] >> PAGE_SHIFT, vma->vm_end - vma->vm_start,
                          vma->vm_page_prot);
  if (rc)
  {
    pr_devel(P40_FMT "io_remap_pfn_range()\n", P40_PARM);
    return rc;
  }
  return 0;
}

static struct file_operations ecs_file_ops = {
    .owner = THIS_MODULE,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35))
    .ioctl = ecs_ioctl,
#else
    .unlocked_ioctl = ecs_ioctl,
#endif
    .mmap = ecs_mmap,
    .open = ecs_open,
    .release = ecs_release,
};

//+`pcie40_ecs_probe` Initialize ECS BARs and create device files.
int pcie40_ecs_probe(struct pci_dev *dev, const struct pci_device_id *id) //;?>
{
  int rc = 0;

  struct pcie40_state *common;
  struct pcie40_ecs_state *state = NULL;

  common = pci_get_drvdata(dev);

  //? This function allocates a p40driver`pcie40_ecs_state` instance to keep ECS-specific state.
  state = kzalloc(sizeof(struct pcie40_ecs_state), GFP_KERNEL);
  if (IS_ERR(state))
  {
    printk(P40_ERR "kzalloc()\n", P40_PARM);
    rc = PTR_ERR(state);
    goto err_kzalloc;
  }
  state->common = common;
  pr_devel(P40_FMT "state = 0x%p\n", P40_PARM, state);

  //? It then requests exclusive access to the ECS BARs
  rc = pci_request_selected_regions_exclusive(dev, P40_ECS_BARS_MASK, P40_DRV_NAME);
  if (rc)
  {
    printk(P40_WARN "unable to reserve ECS regions\n", P40_PARM);
    goto err_pci_request_regions_ecs;
  }

  if (state->common->bar_start[BAR0] && state->common->bar_size[BAR0])
  {
    pr_devel(P40_FMT "pci_iomap() BAR0 (%lu bytes)\n", P40_PARM, state->common->bar_size[BAR0]);
    state->common->bar0_regs = pci_iomap(dev, BAR0, state->common->bar_size[BAR0]);
    if (state->common->bar0_regs == NULL)
    {
      rc = -1;
      printk(P40_ERR "pci_iomap() error \n", P40_PARM);
      goto err_bar0_iomap;
    }
  }

  if (state->common->bar_start[BAR2] && state->common->bar_size[BAR2])
  {
    pr_devel(P40_FMT "pci_iomap() BAR2 (%lu bytes)\n", P40_PARM, state->common->bar_size[BAR2]);
    state->common->bar2_regs = pci_iomap(dev, BAR2, state->common->bar_size[BAR2]);
    if (state->common->bar2_regs == NULL)
    {
      rc = -1;
      printk(P40_ERR "pci_iomap() error \n", P40_PARM);
      goto err_bar2_iomap;
    }
  }

  //? and allocates a range of minor numbers for its character devices.
  rc = alloc_chrdev_region(&(state->dev_num), P40_ECS_CDEV_BASEMINOR, P40_ECS_CDEV_COUNT, P40_DRV_NAME);
  if (rc < 0)
  {
    printk(P40_ERR "alloc_chrdev_region()\n", P40_PARM);
    goto err_alloc_chrdev_region;
  }

  //? One such device is created for BAR0
  if (state->common->bar_size[0])
  {
    rc = pcie40_setup_cdev(pcie40_ecs_class, &(state->bar0_cdev), state->dev_num, BAR0_CDEV_MINOR, 0, BAR0_CDEV_NAME, state->common->dev_id, &ecs_file_ops);
    if (rc < 0)
    {
      printk(P40_ERR "!! BAR0 setup not possible \n", P40_PARM);
      goto err_bar0_dev;
    }
  }
  //? and a second one for BAR2.
  if (state->common->bar_size[2])
  {
    rc = pcie40_setup_cdev(pcie40_ecs_class, &(state->bar2_cdev), state->dev_num, BAR2_CDEV_MINOR, 2, BAR2_CDEV_NAME, state->common->dev_id, &ecs_file_ops);
    if (rc < 0)
    {
      printk(P40_ERR "!! pcie40_setup_cdev not possible  \n", P40_PARM);
      goto err_bar2_dev;
    }
  }

  pcie40_ecs_set_drvdata(dev, state);
  return rc;
err_bar0_dev:
  unregister_chrdev_region(state->dev_num, P40_ECS_CDEV_COUNT);

err_bar2_dev:
  unregister_chrdev_region(state->dev_num, P40_ECS_CDEV_COUNT);
err_alloc_chrdev_region:
  iounmap(state->common->bar2_regs);
err_bar0_iomap:
  pci_release_selected_regions(dev, P40_ECS_BARS_MASK);
err_bar2_iomap:
  pci_release_selected_regions(dev, P40_ECS_BARS_MASK);
err_pci_request_regions_ecs:
  kfree(state);
err_kzalloc:
  return rc;
}

//+`pcie40_ecs_remove` Destroy ECS device files.
void pcie40_ecs_remove(struct pci_dev *dev) //;?>
{
  struct pcie40_ecs_state *state;

  pr_devel(P40_FMT "pci_dev = 0x%p\n", P40_PARM, dev);
  state = pcie40_ecs_get_drvdata(dev);

  if (!dev || !state)
  {
    pr_devel(P40_FMT "remove(dev = 0x%p) dev->driver_data = 0x%p\n", P40_PARM, dev, state);
    return;
  }

  if (state->common->bar_size[2])
  {
    printk(P40_INFO "remove /dev/pcie40_%d_%s\n", P40_PARM, state->common->dev_id, BAR2_CDEV_NAME);
    device_destroy(pcie40_ecs_class, MKDEV(MAJOR(state->dev_num), MINOR(state->dev_num) + BAR2_CDEV_MINOR));
  }
  if (state->common->bar_size[0])
  {
    printk(P40_INFO "remove /dev/pcie40_%d_%s\n", P40_PARM, state->common->dev_id, BAR0_CDEV_NAME);
    device_destroy(pcie40_ecs_class, MKDEV(MAJOR(state->dev_num), MINOR(state->dev_num) + BAR0_CDEV_MINOR));
  }
  unregister_chrdev_region(state->dev_num, P40_ECS_CDEV_COUNT);

  printk(P40_INFO "releasing PCI regions\n", P40_PARM);
  iounmap(state->common->bar2_regs);
  pci_release_selected_regions(dev, 2);

  kfree(state);
}
