// File for Belle 2 DMA functions for PCIe40
#include "pcie40_b2dma.h"

#include "dma_driverlib.h"
#include "pcie40_ecs.h"

#include <stdio.h>
#include <string>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>
#include <errno.h>
#include <cstring>
#include <unistd.h>

// **************************************************************************
// Stop the DMA
// **************************************************************************
int pcie40_dmaStop(int fd) { 
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_CTRL , 0 ) ;
  return ret ;
}

// **************************************************************************
// Start the DMA
// **************************************************************************
int pcie40_dmaStart(int fd) { 
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_CTRL , 0x81 ) ;
  return pcie40_start_dma( fd ) ;
}

// **************************************************************************
// Reset the DMA
// **************************************************************************
int pcie40_dmaReset(int fd) { 
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_RST , 1 << DMA_RST_BIT ) ;
  return ret ;
}

// **************************************************************************
// Set readout DMA
// **************************************************************************
int pcie40_dmaSetReadout( int fd ) {
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_FIFO_0 , 0 ) ;
  return ret ;
}

// **************************************************************************
// Set size from header
// **************************************************************************
int pcie40_setSizeFromHeader( int fd ) {
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_FIFO_1 , 0 ) ;
  return ret ;
}

// **************************************************************************
// Disable size from header
// **************************************************************************
int pcie40_disableSizeFromHeader( int fd ) {
  int ret = 0 ;
  ret = ecs_write( fd , DMA_BAR , DMA_FIFO_1 , 1 ) ;
  return ret ;
}

// **************************************************************************
// Disable data generator
// **************************************************************************
int pcie40_enableGenerator(int fd) { 
  int ret = 0 ;
  int value ; 
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ; 
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( 1 << DISABLE_GENERATOR_BIT ) | value ) ;
  return ret ;
}

// **************************************************************************
// Enable data generator
// **************************************************************************
int pcie40_disableGenerator(int fd) { 
  int ret = 0 ;
  int value ; 
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ; 
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( ~( 1 << DISABLE_GENERATOR_BIT ) ) & value ) ;
  return ret ;
}

// **************************************************************************
// Enable external trigger for data generator
// **************************************************************************
int pcie40_enableExternalTrigger(int fd) { 
  int ret = 0 ;
  int value ; 
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ; 
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , 
		   ( 1 << USE_EXT_TRIG_BIT ) | value ) ;
  return ret ;
}

// **************************************************************************
// Disable external trigger for data generator
// **************************************************************************
int pcie40_disableExternalTrigger(int fd) { 
  int ret = 0 ;
  int value ; 
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ; 
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , 
		   ( ~( 1 << USE_EXT_TRIG_BIT ) ) & value ) ;
  return ret ;
}

// **************************************************************************
// Set Busy level (not sure it is used)
// **************************************************************************
int pcie40_setBusyLevel( int fd , unsigned int level ) {
  return ecs_write( fd , LLI_BAR , DMA_BUSY_LEVEL , level ) ;
}


// **************************************************************************
// Use data from fibers
// **************************************************************************
int pcie40_useDataFromGenerator( int fd ) { 
  int ret = 0 ;
  int value ;
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ;
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( ~( 1 << USE_DATA_FROM_FIBERS_BIT ) ) & value ) ;
  return ret ;
}

// **************************************************************************
// Use free signal
// **************************************************************************
int pcie40_useFreeSignal( int fd ) { 
  int ret = 0 ;
  int value ;
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ;
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( 1 << USE_FREE_BIT ) | value ) ;
  return ret ;
}

// **************************************************************************
// Ignore free signal
// **************************************************************************
int pcie40_ignoreFreeSignal( int fd ) { 
  int ret = 0 ;
  int value ;
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ;
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( ~( 1 << USE_FREE_BIT )) & value ) ;
  return ret ;
}


// **************************************************************************
// Set trigger frequency of data generator
// **************************************************************************
int pcie40_setGeneratorTriggerFrequency( int fd , unsigned int freq ) {
  int ret = 0 ;
  int value ; 
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ; 
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( ( 0xFF & freq ) | ( 0xFF00 & value ) ) ) ;
  return ret ;
}

// **************************************************************************
// Get trigger frequency of data generator
// **************************************************************************
int pcie40_getGeneratorTriggerFrequency( int fd ) {
  int value ;
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ;
  return ( value & 0xFF ) ;
}

// **************************************************************************
// Use data from generator
// **************************************************************************
int pcie40_useDataFromFibers( int fd ) {
  int ret = 0 ;
  int value ;
  value = ecs_read( fd , LLI_BAR , REG_TRIG_FREE_RESET ) ;
  ret = ecs_write( fd , LLI_BAR , REG_TRIG_FREE_RESET , ( 1 << USE_DATA_FROM_FIBERS_BIT ) | value ) ;
  return 0 ;
}

// **************************************************************************
// number of words in event (for data generator)
// **************************************************************************
int pcie40_setNbWordInEvent( int fd , unsigned int n ) {
  int ret = 0 ;
  ret = ecs_write( fd , LLI_BAR , REG_NB_WORD_IN_EVENT , n & 0xFF ) ;
  return ret ;
}

// **************************************************************************
// number of words in event (for data generator)
// **************************************************************************
unsigned int pcie40_getNbWordInEvent( int fd ) {
  return ( ecs_read( fd , LLI_BAR , REG_NB_WORD_IN_EVENT ) & 0xFF ) ;
}

// **************************************************************************
// Number of events (internal data generator)
// **************************************************************************
int pcie40_setNbEvents( int fd , unsigned int n ) {
  int ret = 0 ;
  ret = ecs_write( fd , LLI_BAR , REG_NB_EVENTS , n & 0xFFFFFFFF ) ;
  return ret ;
}

// **************************************************************************
// Get number of events (internal data generator)
// **************************************************************************
unsigned int pcie40_getNbEvents( int fd ) {
  return ecs_read( fd , LLI_BAR , REG_NB_EVENTS ) ;
}

// **************************************************************************
// Reset the address of the DMA buffer
// **************************************************************************
int pcie40_b2dmabufferreset(int fd) { 
  return 0 ;
}

// **************************************************************************
// Resynchronize the b2link, returns 1 if success, 0 if error 
// **************************************************************************
int pcie40_b2dmabufferread(int fd , unsigned int result[] ) { 
  printf( "ERROR: not yet implemented for new DMA\n" ) ;
  return 1 ;
}

// **************************************************************************
// Resynchronize the b2link, returns 1 if success, 0 if error 
// **************************************************************************
int pcie40_b2dmapointerread(int fd , unsigned int ** pData ) { 
  return 1 ;
}


